#
# Copyright(c) 2006-2016 ThreatSTOP, Inc. All Rights Reserved
#
# NOTICE: All information contained herein is, and remains the property
# of ThreatSTOP, Inc. and its suppliers, if any. The intellectual and 
# technical concepts contained herein are proprietary to ThreatSTOP, Inc. 
# and its suppliers and certain aspects thereof are protected by United 
# States Patent No. 8,533,822 and United States Patent No. 8,869,237, 
# and are protected by trade secret or copyright law.
#
# US Government Entities: The ThreatSTOP service, software and
# documentation, as applicable, are "commercial computer software" and
# "commercial computer software documentation" developed exclusively at
# private expense by Threatstop, Inc. ("Threatstop"). Pursuant to FAR
# 12.212 or DFARS 227 7202 and their successors, as applicable, use,
# reproduction and disclosure of the software and documentation is
# governed by the terms of Threatstop's commercial agreement.
#

# file : ts-common.sh
# This is intended to be included by all executable shell scripts in
# this package.
#
# NOTE : this "dots" dns.sh, so other scripts do not have to
#

# all scripts can assume these are set and usable
#
# INSTALL_DIR - determined by where the script was located
# DEBUG - indicates to echoMsg to also output DEBUG level messsages
# AUDIT_LOG - all echoMsg strings are logged here, regardless of $DEBUG
# *_LIST_PATH - qualify the block & allow lists paths under INSTALL_DIR
#
setGlobals() {

    local ORIG=$(pwd)
    cd $(dirname $0) # always where the script was run from
    INSTALL_DIR=$(pwd)
    cd ${ORIG}

    if [ "$1" = "DEBUG" ]; then

        DEBUG=true
    fi

    AUDIT_LOG_DIR=${INSTALL_DIR}/audit
    AUDIT_LOG=${AUDIT_LOG_DIR}/$(basename $0).run.log
    mkdir -p ${AUDIT_LOG_DIR}
    if [ -e ${AUDIT_LOG} ]; then

        cp ${AUDIT_LOG} ${AUDIT_LOG}.bak
        echo -n "" >${AUDIT_LOG}
    fi

    # with the above, *NOW* echoMsg and myExit may be used

    # qualify path to lists
    if [ ! -d ${INSTALL_DIR}/lists ]; then

        mkdir -p ${INSTALL_DIR}/lists 
        if [ $? -ne 0 ]; then

           myExit 1 ERROR "Unable to create lists directory"
        fi
    fi

    . ${INSTALL_DIR}/ts-dns.sh
}

# feedback to the user, and for audit logging
echoMsg() {

    local msg=""
    local dbg=""
    local old_ifs="${IFS}"
    IFS=
    case $1 in
        ERROR) msg="[ERROR] : $2"
        ;;
        WARN)  msg="[WARN ] : $2"
        ;;
        ALERT) msg="[ALERT] : $2"
        ;;
        INFO)  msg="[INFO ] : $2"
        ;;
        DEBUG)
               dbg="    [DEBUG] : $2"
               if [ "$DEBUG" = true ]; then
                   msg=$dbg
               else
                   echo -e $dbg >>${AUDIT_LOG}
               fi
        ;;
        *) msg="SCRIPT ERROR : Unrecognized parameter [$1]"
    esac

    if [ "$msg" != "" ]; then

        echo -e $msg
        echo -e $msg >>${AUDIT_LOG}
    fi
    IFS="${old_ifs}"
}

# every script that uses these utilities should call this immediately after include this file
# returns
# 0 for success lock
# 1 unable to gain a lock; caller should not proceed
setLock() {

    if [ "$LOCKFILE" = "" ]; then

        echoMsg ERROR "LOCKFILE not set; the caller should set this first"
        exit 1
    fi

    echoMsg DEBUG "Locking current instance"
    if [ -e $LOCKFILE ]; then

        ps=$(cat $LOCKFILE)
        echoMsg ERROR "Failed to lock ... Another instance [$ps] of the same script is running"
        return 1
    fi

    # single quote the code snippet, so to evaluate $?
    # at trap time, not at time of trap definition;
    # this will exit with what was intended
    trap 'EXIT_STATUS=$?; /bin/rm -f $LOCKFILE; exit $EXIT_STATUS' 0 1 2 3 15

    echo $$ > $LOCKFILE
    return 0
}

# assumes LOCKFILE is set by the includer of these utilities
# includer should call this before exiting
unLock() {

    echoMsg DEBUG "Unlocking current instance"
    /bin/rm -f $LOCKFILE
}

# common exit point
# call with val level message, like os
# myExit 1 ERROR "some error message"
myExit() {

    local val=$1
    local level=$2
    local msg="$3"

    echoMsg $level "$msg"

    unLock
    exit $val
}

# determineSedProperties()
# bsd and gnu sed have some differences, the most notable
# being how to designate extended regular expression processing
#
# this will early-out, if defined in config file
#
# global set : SED_EXT_REGEXP
#
# returns :
# 0 for one found
# 1 for not found
determineSedProperties() {

    echoMsg DEBUG "determineSedProperties():"

    if [ "${SED_EXT_REGEXP}" != "" -a -x "${SED_EXT_REGEXP}" ]; then

        echoMsg DEBUG "SED_EXT_REGEXP (${SED_EXT_REGEXP}) is set in config file; will not probe for it"
        return 0
    fi

    echo foo | sed -E 'd;' > /dev/null 2>&1
    if [ $? -eq 0 ]; then

        echoMsg DEBUG "Found that -E is appropriate for this sed"
        SED_EXT_REGEXP='-E'
        return 0
    fi

    echo foo | sed -r 'd;' > /dev/null 2>&1
    if [ $? -eq 0 ]; then

        echoMsg DEBUG "Found that -r is appropriate for this sed"
        SED_EXT_REGEXP='-r'
        return 0
    fi

    echoMsg ERROR "Unable to determine all sed properties; should not continue"
    return 1
}

# determineWebTool()
# find where the executable web tool exists, if any
#
# globals set :
#
# - WEB_TOOL
#
# Note, once this is successful, the user can use
# WEB_TOOL.  However, they're responsible for using
# correct flags for the task hand, and the tool that's
# set.
#
# returns :
# 0 for one found
# 1 for not found
determineWebTool() {

    local path
    local paths="/bin /sbin /usr/bin /usr/sbin"
    local tool
    local tools="curl fetch wget"

    echoMsg DEBUG "determineWebTool():"

    if [ "${WEB_TOOL}" != "" -a -x "${WEB_TOOL}" ]; then

        echoMsg DEBUG "WEB_TOOL (${WEB_TOOL}) is set in config file; will not probe for it"
        return 0
    fi

    for tool in ${tools}; do

        for path in ${paths}; do

            if [ -x "${path}/${tool}" ]; then

                echoMsg DEBUG "setting WEB_TOOL to ${path}/${tool}"
                WEB_TOOL="${path}/${tool}"
                return 0
            fi
        done
    done

    echoMsg ERROR "Unable to find one of '${tools}'; should not continue"
    return 1
}

#
# returns
# takes the passed variable name and value and updates
# the main configuration file accordingly (creates it if
# not there, or updates it if exists)
# args :
# - name
# - value
#
# returns
# 0 for success
# 1 for failure (and may output ERROR information)
writeConfigItem() {

    local name=$1
    local value=$2
    local cfg
    local rval
    cfg=${INSTALL_DIR}/threatstop.conf

    echoMsg DEBUG "writeConfigItem():"

    test -e $cfg && grep -q "^${name}=" $cfg
    if [ $? -eq 0 ]; then

        echoMsg DEBUG "Updating ${name}=\"${value}\""
        sed -i.bak ${SED_EXT_REGEXP} -e "s/^${name}=\"?.+\"?$/${name}=\"${value}\"/;" $cfg
    else

        echoMsg DEBUG "Adding ${name}=\"${value}\""
        echo "${name}=\"$value\"" >> $cfg
    fi

    rval=$?    
    if [ $rval -eq 0 ]; then

        echoMsg DEBUG "Success"
    else

        echoMsg DEBUG "Failure"
    fi

    return $rval
}

# loadConfig()
# populate the well-known variables, as follows
# 0. BAIL, if CONFIG_VERSION from config file doesn't match this
#    software version
# 1. load threatstop.conf (if it exists...if not, bail, since
#    the customer has yet to follow the cut-n-paste instructions
#    from the portal)
#    TSCM note : this will always exist
# 2. load threatstop.conf.template
#    This brings in default values, if needed (see next)
# 3. Examine critical variables, loading defaults from the template
#    if they do not yet exist.
# 4. One-off probes, like SRV, current system model, version, device IP, ...
#
# returns
# 0 for success
# 1 for failure
loadConfig() {

    local cfg=${INSTALL_DIR}/threatstop.conf
    local tpl=$cfg.template
    local target
    local old_ifs
    local srv_url

    echoMsg DEBUG "loadConfig():"

    # load version
    TS_VERSION=$(cat ${INSTALL_DIR}/VERSION)
    echoMsg DEBUG "ThreatSTOP version ${VERSION}"

    if [ -e $cfg ]; then

         . $cfg # load all existing variables
    else
         echoMsg WARN "No prior configuration found; starting with default values"
    fi

    if [ "${CONFIG_VERSION}" = "" ]; then

        # initial install, so set current version in config file
        writeConfigItem CONFIG_VERSION ${TS_VERSION} || myExit 1 ERROR "Unable to set CONFIG_VERSION to ${TS_VERSION} in the configuration file"
    else

        # see if they're running new software on old config
        if [ "${CONFIG_VERSION}" != "${TS_VERSION}" ]; then

             echoMsg ERROR "You are attempting to run version ${TS_VERSION} software with a ${CONFIG_VERSION} configuration."
             echoMsg ERROR "You must first run \"ts-upgrade\"; this will perform required modifications."
             myExit 1 ERROR "Exiting..."
        fi
    fi

    if [ ! -e $tpl ]; then

         echoMsg ERROR "Cannot find $tpl; this should have been installed with this package"
         return 1
    fi

    # OPTIONAL TEMPLATE SUPPORT
    # For variables in the template file...the user of this script
    # *must* supply a threatstop.conf.template file, and it
    # *may* contain variables for the following code to process.
    # The processing will check for variable content, loading from
    # template value if there is no content; ie, an empty string.
    # This will know to process a variable by referencing the
    # TEMPLATE_VARS variable, provided in the template file.
    #
    # Mandatory interface to use this processing :
    # The template file must provide variable T, which is a prefix
    # string to each variable name.  The variables in the template
    # file must contain that prefix.  For example, if it is desired
    # to have FOO and BAR processed by the following loop, this
    # is an example of a compatible template file :
    #
    # TEMPLATE_VARS="FOO BAR"
    # T=myprefix_
    # myprefix_FOO="foo default value"
    # myprefix_BAR="bar default value"
    #
    if [ -e $tpl ]; then

        . $tpl # load all template variables, as well as the template prefix $T
    
        if [ "${TEMPLATE_VARS}" != "" ]; then

            local var
            local ivar
            local tvar
            for var in ${TEMPLATE_VARS}; do
        
                eval ivar=\$${var} # indirect value
                if [ "${ivar}" = "" ]; then
        
                    eval tvar=\$${T}${var} # get template value
                    eval "${var}=${tvar}"
                fi
            done
            BLOCK_LIST_PATH=${INSTALL_DIR}/${BLOCK_LIST_PATH}
            ALLOW_LIST_PATH=${INSTALL_DIR}/${ALLOW_LIST_PATH}
            # all template variables should be available now
        fi
    fi

    # see if we have dig, host, or other
    determineSedProperties
    if [ $? -ne 0 ]; then

        echoMsg ERROR "Will not continue without sed properties"
        exit 1
    fi

    # see if we have dig, host, or other
    determineLookupTool
    if [ $? -ne 0 ]; then

        echoMsg ERROR "Have no way to perform DNS lookups; exiting"
        exit 1
    fi

    # see if we have a web utility
    determineWebTool
    if [ $? -ne 0 ]; then

        echoMsg ERROR "Have no way to perform Web lookups; exiting"
        exit 1
    fi

    # see if we need to update the DNS server list
    lookupSRV tsdns tcp threatstop.com
    if [ $? -eq 0 ]; then

        echoMsg DEBUG "DNS server SRV : ${SRV_OUT}"

        target=$(echo ${SRV_OUT} | awk '{print $4;}' | sed 's/\.$//;')
        if [ "${target}" != "${DNS_SERVER_LIST}" ]; then

            echoMsg DEBUG "Using ${target} for policy lookups"
            DNS_SERVER_LIST=${target}
            updateConfig DNS_SERVER_LIST ${DNS_SERVER_LIST}
        fi
    fi


    # see if we need to update the FTP upload URL
    lookupSRV logs sftp threatstop.com
    if [ $? -eq 0 ]; then

        echoMsg DEBUG "FTP URL SRV : ${SRV_OUT}"

        target=$(echo ${SRV_OUT} | awk '{print $4;}' | sed 's/\.$//;')
        srv_url="ftp://ftp@${target}/logs"
        if [ "${srv_url}" != "${FTP_URL}" ]; then

            echoMsg DEBUG "Will ftp logs to ${srv_url}"
            FTP_URL=${srv_url}
            updateConfig FTP_URL ${FTP_URL}
        fi
    fi

    echoMsg DEBUG "Checking if log upload IP needs probing for"
    if [ "${LOG_UPLOAD_IP}" = "" ]; then

        # Most of the time, the correct log upload IP is the IP as
        # seen by the ThreatSTOP servers.  This probe will determine
        # that IP, and prime LOG_UPLOAD_IP.
        checkValidIP
        if [ "${DEVICE_IP_PROBED}" != "" ]; then

            echoMsg INFO "Defaulting the log association IP to ${DEVICE_IP_PROBED}"
            LOG_UPLOAD_IP=${DEVICE_IP_PROBED}
            writeConfigItem LOG_UPLOAD_IP ${LOG_UPLOAD_IP}
        else

            echoMsg ERROR "The ThreatSTOP servers either cannot be reached, or the IP"
            echoMsg ERROR "of this device is not allowed to query the ThreatSTOP servers"
            echoMsg ERROR "...correct the network configuration and/or the portal configuration"
            return 1
        fi
    fi

    return 0
}

# updateConfig()
# set the named variable to the value provided; if it already
# exists, the value will be updated
#
# args
# - variable name
# - variable value
#
# returns
# 0 for success
# 1 for failure
updateConfig() {

    local cfg=${INSTALL_DIR}/threatstop.conf
    local name=$1
    local value="$2"

    echoMsg DEBUG "updateConfig():"

    if [ -e "${cfg}" ]; then

        grep -q "^${name}=" $cfg
        if [ $? -eq 0 ]; then

            echoMsg DEBUG "updating $name to '$value'"
            sed -i.bak -e "s|^${name}=.*$|${name}=\"${value}\"|;" $cfg
        else

            echoMsg DEBUG "creating $name with '$value'"
            echo "${name}=\"${value}\"" >> $cfg
        fi
    else

        echoMsg DEBUG "creating $cfg"
        echoMsg DEBUG "creating $name with '$value'"
        echo "${name}=\"${value}\"" >> $cfg
    fi

    return $?
}

# simple sanity checks used
# take the item to check
# return
# 0 for success
# 1 for failure
confirmPolicy() {
    echo "$1" | grep -q -E '[^-.a-zA-Z0-9]' # looking for invalid chars
    if [ $? -eq 0 ]; then

        return 1
    else

        queryPolicyZone "$1"
        if [ $? -eq 0 ]; then

            return 0
        else

            return 1
        fi
    fi
}
confirmIP() {
    echo "$1" | grep -q -E '^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$' # positive match
    if [ $? -eq 0 ]; then

        return 0
    else

        return 1
    fi
}
confirmInt() {
    echo "$1" | grep -q -E '^[0-9]+$' # positive match
    if [ $? -eq 0 ]; then

        return 0
    else

        return 1
    fi
}

# Make sure the IP this device is seen as by ThreatSTOP
# is allowed to query.  The customer's portal configuration
# for this device drives ACLs that allow this IP.
#
# globals used :
# - WEB_TOOL
# - VALID_IP_URL
# globals set :
# - DEVICE_IP_PROBED is set with the IP as seen by the check CGI
#
# returns :
# 0 success
# 1 failure
checkValidIP() {

    local rval=0
    local valid_ip_out
    local valid_ip_ip
    local timeout=10
    local tool
    local tool_args

    case $(basename ${WEB_TOOL}) in
        curl) tool_args="--silent --connect-timeout ${timeout} ${VALID_IP_URL}"
        ;;
        fetch) tool_args="-T ${timeout} -q -o - ${VALID_IP_URL}"
        ;;
        wget)  tool_args="-T ${timeout} -q -O - ${VALID_IP_URL}"
        ;;
        *)  echoMsg ERROR "Cannot find a web tool with which to probe ${VALID_IP_URL}"
            return 1
        ;;
    esac

    valid_ip_out=$(${WEB_TOOL} ${tool_args})
    if [ $? -eq 0 ]; then

        valid_ip_ip=$(echo "${valid_ip_out}" | sed -n ${SED_EXT_REGEXP} 's/^Your IP address: ([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)$/\1/p;')
        echoMsg DEBUG "Query came from '${valid_ip_ip}'"

        echo "${valid_ip_out}" | grep -q 'Address is in the list of authorized hosts'
        if [ $? -eq 0 ]; then

            echoMsg DEBUG "${valid_ip_ip} is okay to query ThreatSTOP"
        else

            echoMsg DEBUG "${valid_ip_out}"
            rval=1
        fi
    else

        echoMsg ERROR "Attempt to reach ThreatSTOP validation server failed"
        echoMsg ERROR "(${VALID_IP_URL})"
        rval=1
    fi

    DEVICE_IP_PROBED=${valid_ip_ip}

    return $rval
}

# these get run when this utility code is "dotted"
setGlobals $1
# this is now handled; let's cleanup the command line
if [ "$1" = "DEBUG" ]; then

    echoMsg DEBUG "Removing DEBUG from command-line args"
    shift 1
fi
